﻿using System;
using System.Collections.Generic;

using PT.Scheduler;
using PT.Scheduler.Simulation.Customizations;
using PT.Transmissions;

namespace Planettogether.Sample.CampaignPostProcessing
{
    /// <summary>
    /// 
    /// </summary>
    public class EndOfSimulation : EndOfSimulationCustomization
    {
        public override string Name => "EndOfSimulation";

        public override string Description => "Triggers a time adjustment simulation after an optimize simulation has completed.";

        private bool m_runningTimeAdjustment = false;

        private RequiredCapacityCust m_requiredCapacityCust = null;
        private Dictionary<BaseId, TimeSpan> m_activityPostProcessing = new Dictionary<BaseId, TimeSpan>();

        // the name of the attribute whose code determines which activities
        // are in the same campaign. Also, the attribute should specify
        // the duration of campaign post processing in the setuphrs property.
        private static string CampaignAttributeName = "Attribute1";

        internal static Tuple<string, TimeSpan> GetCampaignAttributes(InternalActivity a_act)
        {
            PTAttribute attr = a_act.Operation.Attributes.Find(EndOfSimulation.CampaignAttributeName);
            return new Tuple<string, TimeSpan>(attr.Code, attr.SetupSpan);
        }

        private void BroadcastTimeAdjustment(ScenarioDetail a_sd, ScenarioBaseT a_t)
        {
            if (!PTSystem.Server) return;

            ScenarioTouchT touchT = new ScenarioTouchT();


            using (var oa = a_sd.Scenario.ScenarioSummaryLock.EnterRead())
            {
                touchT.LastTransmissionNbr = oa.Instance.LastTransmissionNbr;
            }
            touchT.sequenced = false;
            touchT.TransmissionSender = PTTransmissionBase.TransmissionSenderType.PTSystem;

            this.AddTransmissionsToProcess(1, touchT);
        }


        protected override void SimulationInitialization(ScenarioDetail a_sd, ScenarioDetail.SimulationType a_simulationType, ScenarioBaseT a_transmission)
        {
            base.SimulationInitialization(a_sd, a_simulationType, a_transmission);

            //Restore reference to other customization points if needed
            if (m_requiredCapacityCust == null)
            {
                m_requiredCapacityCust = a_sd.CustomizationInstances.SchedulerCustomization as RequiredCapacityCust;
                m_requiredCapacityCust.SetActivityPostProcessing(m_activityPostProcessing);
            }

            //Remove previous extra transmissions
            CleanTransmissionsProcess();
        }

        private void DetermineActivityPostProcessing(Resource a_res)
        {
            ResourceBlockList.Node prev = a_res.Blocks.First;
            while(prev != null)
            {
                ResourceBlockList.Node next = prev.Next;
                if (next != null)
                {
                    InternalActivity prevAct = prev.Data.Batch.FirstActivity;
                    var prevActCampaignAttrs = GetCampaignAttributes(prevAct);
                    var nextActCampaignAttrs = GetCampaignAttributes(next.Data.Batch.FirstActivity);
                    if (prevActCampaignAttrs.Item1 != nextActCampaignAttrs.Item1)
                    {
                        m_activityPostProcessing.Add(prevAct.Id, prevActCampaignAttrs.Item2);
                    }
                }

                prev = next;
            }
        }

        private void DetermineActivityPostProcessing(ScenarioDetail a_sd)
        {
            m_activityPostProcessing.Clear();

            foreach(Plant p in a_sd.PlantManager)
            {
                foreach(Department dept in p.Departments)
                {
                    foreach (Resource res in dept.Resources)
                    {
                        DetermineActivityPostProcessing(res);
                    }
                }
            }
        }

        protected override void EndOfSimulationCleanup(ScenarioDetail.SimulationType a_simType, ScenarioBaseT a_t, ScenarioDetail a_sd)
        {
            base.EndOfSimulationCleanup(a_simType, a_t, a_sd);
            if (a_simType == ScenarioDetail.SimulationType.Optimize)
            {
                m_runningTimeAdjustment = true;
                DetermineActivityPostProcessing(a_sd);
                BroadcastTimeAdjustment(a_sd, a_t);
            }
            else if (a_simType == ScenarioDetail.SimulationType.TimeAdjustment && m_runningTimeAdjustment)
            {
                m_activityPostProcessing.Clear();
                m_runningTimeAdjustment = false;
            }
        }
    }
}
