﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using PT.Scheduler;

namespace Planettogether.Sample.CampaignPostProcessing
{
    public class Schedulability : PT.Scheduler.Simulation.Customizations.SchedulabilityCustomization
    {
        public override string Name => "Campaign Post Processing";

        public override string Description => "Add post processing time after campaign end. This part of the customization will push out campaigns to make room for post processing";

        /**
         * returns the last activity scheduled on the resource or null if 
         * no activities are scheduled on the resource.
         */
        private InternalActivity GetLastScheduledActivity(Resource a_res)
        {
            return a_res.Blocks.Last?.Data.Batch.FirstActivity;
        }

        protected override bool IsSchedulable(ScenarioDetail a_sd, ScenarioDetail.SimulationType a_simType, long a_clock, long a_simulationClock, BaseResource a_res, BaseActivity a_activity, SchedulableInfo a_schedulableInfo, out long o_postEventTime)
        {
            o_postEventTime = -1;

            // only apply this add-in logic during optimize.
            if (a_simType != ScenarioDetail.SimulationType.Optimize)
            {
                return true;
            }

            InternalActivity act = a_activity as InternalActivity;
            Resource res = a_res as Resource;

            InternalActivity lastScheduledAct = GetLastScheduledActivity(res);
            if (lastScheduledAct == null)
            {
                // this is the first activity being scheduled on the resource. allow it to schedule.
                return true;
            }

            var lastActCampaignAttrs = EndOfSimulation.GetCampaignAttributes(lastScheduledAct);
            var currActCampaignAttrs = EndOfSimulation.GetCampaignAttributes(act);
            if (lastActCampaignAttrs.Item1 == currActCampaignAttrs.Item1)
            {
                // same campaign
                return true;
            }

            // activity attempting to schedule is not in the same campaign. Need to delay its scheduling
            // to ensure there's enough time for post processing to be added to the last campaign.

            // calculate available capacity between the last scheduled act end date and 
            // a_simulationClock. a_simulationClock is the datetime that act is attempting to schedule to start at.
            TimeSpan available = res.ResourceCapacityIntervals.GetTotalSchedulableCapacityBetweenDates(lastScheduledAct.ScheduledEndDate, new DateTime(a_simulationClock));

            // there's enough availability for post processing to be added to the end of last act.
            if (available >= lastActCampaignAttrs.Item2)
            {
                return true;
            }

            // we can use o_postEventTime to suggest to the schedule engine the next time the activity should attempt to schedule.
            // in our case, we'll set this to a_simulationClock + how much more time we need to create the gap required.
            o_postEventTime = a_simulationClock + (lastActCampaignAttrs.Item2 - available).Ticks;
            return false;
        }
    }
}
